from bellastore.database.db import Db
import argparse

def main():
    cli = argparse.ArgumentParser()
    cli.add_argument(
        '--root_dir', type = str, default = '/data/deep-learning/test',
       help = 'Directory where sqlite and storage will be initialized under, in particular root_dir/storage/scans.sqlite'
    )
    cli.add_argument(
        '--ingress_dir', type = str, default = '/data/deep-learning/test/slides',
        help = 'Directory holding scans to be inserted into storage.'
    )
    cli.add_argument(
        '--sqlite_name', type = str, default = 'scans.sqlite',
        help = 'Name of the sqlite database file'
    )
    cli.add_argument(
        '--verbose', action=argparse.BooleanOptionalAction,
        help = 'This will log the whole content of the database and the storage which might clutter the output'
    )
    cli.add_argument(
        '--move', action=argparse.BooleanOptionalAction,
        help = 'This needs to be explicitly set in order to mess with the filesystem, otherwise only dry run will be done.'
    )

    args = cli.parse_args()
    root_dir = args.root_dir
    ingress_dir = args.ingress_dir
    sqlite_name = args.sqlite_name
    verbose = args.verbose
    move = args.move
    

    db = Db(root_dir, ingress_dir, sqlite_name)

    if move:
        db.insert_from_ingress()
        if verbose:
            print('Done, your final storage looks like:')
            print(str(db))
    else:
        db = Db(root_dir, ingress_dir, sqlite_name)
        if verbose:
            print('The storage currently looks like this')
            print(str(db))
        valid_scans = db.get_valid_scans_from_ingress()
        formats = set([scan.filename.split('.')[1] for scan in valid_scans])
        print(formats)
        print(f'Total amount of valid scans to be moved {len(valid_scans)}.')
        print(f'Formats present {formats}')


if __name__ == '__main__':
    main()


from bellastore.database.db import Db
import argparse

def main():
    cli = argparse.ArgumentParser()
    cli.add_argument(
        '--root_dir', type = str, default = '/data/deep-learning/storage',
       help = 'Directory where sqlite and storage will be initialized under, in particular root_dir/storage/scans.sqlite'
    )
    cli.add_argument(
        '--sqlite_name', type = str, default = 'scans.sqlite',
        help = 'Name of the sqlite database file'
    )
    args = cli.parse_args()
    root_dir = args.root_dir
    sqlite_name = args.sqlite_name

    db = Db(root_dir=root_dir, ingress_dir=None, filename=sqlite_name)

    db.setup_logging()
    db.create_backup()


if __name__ == '__main__':
    main()


# This test module holds the integration tests

import os
from os.path import join as _j
from pathlib import Path
from typing import List
import sqlite3

from bellastore.utils.scan import Scan
from bellastore.database.db import Db
from conftest import get_files

def get_files_dirs(dir):
    files = Path(dir).rglob("*")
    file_paths = list(files)
    file_paths = {str(file) for file in file_paths}
    return file_paths

def get_scans(dir):
    scans = []
    files = get_files(dir)
    for file in files:
        scan = Scan(file)
        if not scan.is_valid():
            continue
        scan.hash_scan()
        print(scan.hash)
        scans.append(scan)
    return scans


def get_scan_entry(sqlite_path, scan: Scan, table_name: str):
    with sqlite3.connect(sqlite_path) as conn:
        cursor = conn.cursor()
        cursor.execute(f"SELECT * FROM {table_name} WHERE hash = '{scan.hash}'")
        data = cursor.fetchall()
    return data
    

# Fs checks
def check_ingress(db: Db, scans: List[Scan]):
    files_in_ingress = get_files(db.ingress_dir)
    for scan in scans:
        assert {scan.path}.issubset(files_in_ingress)

def check_storage(db: Db, scans: List[Scan]):
    files_in_storage = get_files(db.storage_dir)
    print(files_in_storage)
    for scan in scans:
        assert {scan.path}.issubset(files_in_storage)
    assert {db.sqlite_path}.issubset(files_in_storage)

def check_not_in_storage(db: Db, scans: List[Scan]):
    files_in_storage = get_files(db.storage_dir)
    for scan in scans:
        assert not {scan.path}.issubset(files_in_storage)

def check_empty_ingress(db: Db):
    assert get_files_dirs(db.ingress_dir) == set()

# Db checks
def check_ingress_db(db: Db, scan: Scan):
    # as in ingress we allow for duplicate hashes, issubset is enough
    assert {tuple((scan.hash, _j(db.ingress_dir, scan.filename) , scan.filename))}.issubset(set(get_scan_entry(db.sqlite_path, scan, 'ingress')))

def check_ingress_db_subfolder(db: Db, scan: Scan):
    # as in ingress we allow for duplicate hashes, issubset is enough
    assert {tuple((scan.hash, _j(db.ingress_dir, scan.scanname, scan.filename) , scan.filename))}.issubset(set(get_scan_entry(db.sqlite_path, scan, 'ingress')))

def check_storage_db(db: Db, scans: List[Scan]):
    # in storage we need strict equality
    for scan in scans:
        assert [(scan.hash, scan.path, scan.filename, scan.scanname)] == get_scan_entry(db.sqlite_path, scan, 'storage')



# MAIN TESTS

# Starting from empty database
def test_empty(root_dir, ingress_dir):
    db = Db(root_dir, ingress_dir, 'scans.sqlite')
    # as the scans will be moved to storage we need to record their original paths
    check_ingress(db, get_scans(db.ingress_dir))
    scans = db.insert_from_ingress()

    # Ingress checks
    for scan in scans:
        check_ingress_db(db, scan)
    check_empty_ingress(db)

    # Storage checks
    check_storage(db, scans)
    check_storage_db(db, scans)

    print(str(db))   

# Starting from already filled database
def test_classic(root_dir, ingress_dir, classic_db):
    # db will be equal to classic_db
    # classic db already holds 2 scans (also in storage)
    db = Db(root_dir, ingress_dir, 'scans.sqlite')
    storage_scans = get_scans(_j(root_dir,'storage'))
    check_ingress(db, get_scans(db.ingress_dir))
    scans = db.insert_from_ingress()

    # Ingress checks
    for scan in scans:
        check_ingress_db(db, scan)
    check_empty_ingress(db)

    # Storage checks
    for scan in scans:
        if scan.path:
            storage_scans.append(scan)
        else:
            check_not_in_storage(db, [scan])

    check_storage(db, storage_scans)
    check_storage_db(db, storage_scans)

    print(str(db))

def test_empty_from_subfolder(root_dir, ingress_dir_with_subfolders):
    db = Db(root_dir, ingress_dir_with_subfolders, 'scans.sqlite')
    # as the scans will be moved to storage we need to record their original paths
    check_ingress(db, get_scans(db.ingress_dir))
    scans = db.insert_from_ingress()

    # Ingress checks
    for scan in scans:
        check_ingress_db_subfolder(db, scan)
        {_j(db.ingress_dir, scan.scanname, f"slide_{scan.filename.split('_')[1]}.sqlite")}.issubset(get_files_dirs(db.ingress_dir))

    # Storage checks
    check_storage(db, scans)
    check_storage_db(db, scans)

    print(str(db)) 

# Starting from already filled database
def test_classic_subfolder(root_dir, ingress_dir_with_subfolders, classic_db_subfolders):
    # db will be equal to classic_db_subfolders
    # classic db already holds 2 scans (also in storage)
    db = Db(root_dir, ingress_dir_with_subfolders, 'scans.sqlite')
    storage_scans = get_scans(_j(root_dir,'storage'))
    check_ingress(db, get_scans(db.ingress_dir))
    scans = db.insert_from_ingress()

    # Ingress checks
    for scan in scans:
        check_ingress_db_subfolder(db, scan)
        {_j(db.ingress_dir, scan.scanname, f"slide_{scan.filename.split('_')[1]}.sqlite")}.issubset(get_files_dirs(db.ingress_dir))

    # Storage checks
    for scan in scans:
        if scan.path:
            storage_scans.append(scan)
        else:
            check_not_in_storage(db, [scan])

    check_storage(db, storage_scans)
    check_storage_db(db, storage_scans)

    print(str(db))




    







import os
from os.path import join as _j
from pathlib import Path
from typing import List
import sqlite3

from bellastore.utils.scan import Scan
from bellastore.database.db import Db
from conftest import get_files

def test_fs(root_dir, ingress_dir):
    db = Db(root_dir, ingress_dir, 'scans.sqlite')
    valid_scan_paths = {scan.path for scan in db.get_valid_scans_from_ingress()}
    assert valid_scan_paths == get_files(ingress_dir)

def test_fs_with_subfolders(root_dir, ingress_dir_with_subfolders):
    db = Db(root_dir, ingress_dir_with_subfolders, 'scans.sqlite')
    valid_scan_paths = {scan.path for scan in db.get_valid_scans_from_ingress()}
    assert valid_scan_paths.issubset(get_files(ingress_dir_with_subfolders))




import os
import glob
import hashlib
import base64
import shutil
from typing import List

from .constants import scan_extensions


class Scan():
    """
    Class representing a single scan.

    Init:
    -----
        **path** _str_ : full path to the scan

    Attributes:
        path (str): the full path to the scan, including the filename and ending
        filename (str): just the filename, without the extension
        hash (str | None, default = None): the hash of the file, empty per default

    Methods
    -------
    <p>
        **get_filename**<em>(self, path) -> str</em><br>constructs the filename out of the full path<br>
        **is_valid**<em>(self) -> bool</em><br>checks if a given file file has a scanner-file ending<br>
        **hash_scan**<em>(self) -> str | None</em><br>creates an unique hash for a scan using `sha256`
    </p>
    """
    def __init__(self, path : str):
        self.path = path
        self.scanname = self.get_scanname(path = self.path)
        self.filename = self.get_filename(path = self.path)
        self.hash : None | str = None


    def get_scanname(self, path : str) -> str:
        """
        Returns the filename without the extension of a given path.

        Args:
            path (str): full path to the file

        Returns:
            filename (str): the filename without extension
        """
        return os.path.splitext(os.path.basename(path))[0]
    # TODO: extend this for mxrs, currently this only works for files
    def move(self, target_dir):
        '''
        Moves a scan file into the target directory
        '''
        try:
            source_path = self.path
            # It is crucial to create the target dir before shutil.move
            os.makedirs(target_dir, exist_ok = True)
            shutil.move(self.path, target_dir)
            self.path = os.path.join(target_dir, self.filename)
            print(f"Successfully moved {source_path} into {self.path}")
        except Exception as e:
            raise RuntimeError(f"File can not be moved from {self.path} into {self.path} due to: {e}")

    
    def get_filename(self, path : str) -> str:
        """
        Returns the filename without the extension of a given path.

        Args:
            path (str): full path to the file

        Returns:
            filename (str): the filename without extension
        """
        return os.path.basename(path)


    def is_valid(self) -> bool:
        """
        Checks if the slide has a scanner-file ending.
        These are `.mrxs`, `.svs`, `.ndpi` and `.tif`.

        Returns:
            is_valid (bool): bool indicating if the scan's path has a valid ending
        """
        #scan_extensions = [".ndpi", ".svs", ".mrxs", ".tif"] # CARE
        return any(self.path.endswith(ending) for ending in scan_extensions)

    # TODO: Lukas integrate and test for mxrs, more modular would also be nicer, e.g. _create_raw_hash,_hash_mxrs ...
    def hash_scan(self) -> str | None:
        """
        Creates an url-safe, base64, utf-8 encoded hash for a scan.
        For scans that consist of more than a single file it hashes the whole directory.
        For non-hashable files it will return `None`.

        Returns:
            hash (str | None): the scan's hash (if non-hashable this is `None`)
        """
        def hash_file(path) -> bytes:
            """
            Hashes a single file in chunks of 64kb
            """
            hash = hashlib.sha256()
            if not os.path.isfile(path):
                raise ValueError(f"{path} is not a file")
            with open(path, "rb") as f:
                while True:
                    data = f.read(65536)
                    if not data:
                        break
                    hash.update(data)
            return hash.digest()

        # Check if the slide even is hashable
        if not self.is_valid():
            print(f"Slide is not valid and thus can not be hashed.")
            return None
        is_mrxs = self.path.endswith(".mrxs")
        if not is_mrxs:
            raw_hash = hash_file(self.path)
            hash = base64.urlsafe_b64encode(raw_hash).decode("utf-8")
            self.hash = hash
            return base64.urlsafe_b64encode(raw_hash).decode("utf-8")

        # For `.mrxs` files check if they are structured correctly
        (mrxs_folder, _) = os.path.splitext(self.path)
        if not os.path.isdir(mrxs_folder):
            print(f"Invalid slide {self.path}: {mrxs_folder} is not a directory")
            return None

        # For `.mrxs` files hash all the files one by one
        hash = hashlib.sha256()
        for root, _, files in os.walk(mrxs_folder):
            for file in files:
                file_path = os.path.join(root, file)
                raw_hash = hash_file(file_path)
                hash.update(raw_hash)
        hash.update(hash_file(self.path))
        hash = base64.urlsafe_b64encode(hash.digest()).decode("utf-8")
        self.hash = hash
        return hash




    def __repr__(self) -> str:
        return f"\nCurrent Path: {self.path}\nCurrent Filename: {self.filename}"


import pytest
from os.path import join as _j
from pathlib import Path
from bellastore.utils.scan import Scan


# Helpers
def create_scan_file(dir, filename: str):
    path = _j(dir, filename)
    open(path, 'w').close()
    return path

def get_all_files(dir):
    files = Path(dir).rglob("*")
    file_paths = list(files)
    file_paths = [str(file_path) for file_path in file_paths]
    return file_paths

# Fixtures in this module's scope (rest is in conftest)
@pytest.fixture(scope="function")
def ndpi_scan_path(root_dir):
    return create_scan_file(root_dir, 'test_scan.ndpi')

@pytest.fixture(scope="function")
def txt_scan_path(root_dir):
    return create_scan_file(root_dir, 'test_scan.txt')

@pytest.fixture(scope="function")
def ndpi_scan(root_dir):
        p = create_scan_file(root_dir, 'test_scan.ndpi')
        return Scan(str(p))

@pytest.fixture(scope="function")
def target_dir(root_dir):
    sub = root_dir / "sub"
    sub.mkdir()
    return sub


def test_scan(ndpi_scan_path):
    """Tests the Scan initialization"""
    scan = Scan(path = ndpi_scan_path)
    assert scan.path == ndpi_scan_path
    assert scan.filename == 'test_scan.ndpi'
    assert scan.scanname == 'test_scan'

def test_is_valid_scan(ndpi_scan_path, txt_scan_path):
    """Test if a valid scan is valid"""
    scan = Scan(path = ndpi_scan_path)
    invalid_scan = Scan(path = txt_scan_path)
    assert scan.is_valid()
    assert not invalid_scan.is_valid()

def test_hashing(ndpi_scan_path):
    """Test if the hasing works"""
    scan = Scan(path = ndpi_scan_path)
    hash = scan.hash_scan()
    print(hash)
    assert hash

def test_move_scan(ndpi_scan, target_dir):
    ndpi_scan.move(target_dir)
    assert [ndpi_scan.path] == get_all_files(target_dir)



import os
from os.path import join as _j
from pathlib import Path
from typing import List

from bellastore.utils.scan import Scan

# blueprint fs
class Fs:
    ''' 
    A class representing a simple file system holding:
    - a storage directory
    - an ingress directory
    - a backup directory (for backing up the database)

    Attributes
    ----------
    root_dir: str
        The root of the fs
    ingress_dir: str | None
        The directory holding new candidate scans to be moved to storage.
        If none there we use the class just for inspecting the current state of the fs
    storage_dir: str
        The directory holding all already recorded scans
    backup_dir: str
        The directory holding database backups
    
    Methods
    -------
    _add_scan_to_ingress:
        Method that adds a scan to the ingress. In the sense of the fs the scan is already
        within the ingress so adding means hashing.
    add_scan_to_storage:
        Method to move a scan from the ingress to storage within the file system.
        The recording in the databse will be handled by the Db class
    get_valid_scans_from_ingress:
        Method that scans the ingress directory for valid scans
    remove_empty_folders:
        Method to remove empty folders, resulting from moving scans to storage
    '''

    def __init__(self, root_dir, ingress_dir: None|str):
        self.root_dir = root_dir
        self.ingress_dir = ingress_dir
        self.storage_dir = _j(root_dir, "storage")
        os.makedirs(self.storage_dir, exist_ok = True)
        # we also provide the option just to look at our fs/db without inserting from an ingress
        self.ingress_dir = ingress_dir
        if self.ingress_dir:
            os.makedirs(self.ingress_dir, exist_ok = True)
        self.backup_dir = _j(root_dir, "backup")
        os.makedirs(self.backup_dir, exist_ok=True)

    @staticmethod
    def _get_files(dir):
        files = Path(dir).rglob("*")
        file_paths = list(files)
        file_paths = [str(file) for file in file_paths if file.is_file()]
        return file_paths
    
    def get_valid_scans_from_ingress(self) -> List[Scan]:
        '''
        Method that scans the ingress directory for valid scans
        '''

        scans = []
        print(f'Reading files from ingress directory {self.ingress_dir}')
        files = self._get_files(self.ingress_dir)
        for file in files:
            scan = Scan(file)
            if not scan.is_valid():
                # print(f'Non-valid {scan.path}')
                continue
            print(f'Valid scan: {scan.path}')
            scans.append(scan)
        return scans

    
    def _add_scan_to_ingress(self, scan: Scan):
        '''
        Method that adds a scan to the ingress, i.e. hashes the scan.
        '''

        # Moving to ingress is equivalent to hashing
        scan.hash_scan()
    def add_scan_to_storage(self, scan: Scan):
        '''
        Main function moving scans from ingress to storage
        '''
        
        # self._add_scan_to_ingress(scan)
        target_dir = os.path.join(self.storage_dir, scan.hash)
        scan.move(target_dir)
    def _add_scans_to_ingress(self, scans: List[Scan]):
        for scan in scans:
            self.add_scan_to_ingress(scan)
    def add_scans_to_storage(self, scans: List[Scan]):
        for scan in scans:
            self.add_scan_to_storage(scan)
    
    def get_files_from_ingress(self):
        print(f"The files in {self.ingress_dir} are")
        return self._get_files(self.ingress_dir)
    def get_files_from_storage(self):
        print(f"The files in {self.storage_dir} are")
        return self._get_files(self.storage_dir)
    
    def remove_empty_folders(self):
        '''
        Recursively remove empty folders, resulting from moving scans to storage.
        '''
        # Walk through directory tree in bottom-up order
        for root, dirs, files in os.walk(self.root_dir, topdown=False):
            for dir_name in dirs:
                full_path = os.path.join(root, dir_name)
                if full_path == self.backup_dir:
                    continue
                try:
                    # If directory is empty, remove it
                    if not os.listdir(full_path):
                        os.rmdir(full_path)
                        print(f"Removed empty folder: {full_path}")
                except OSError as e:
                    print(f"Error removing {full_path}: {e}")

    


    def print_tree(self, path=None, prefix=''):
        if path is None:
            path = Path(self.root_dir)
        else:
            path = Path(path)
        
        contents = sorted(path.iterdir(), key=lambda p: p.name)
        
        for i, item in enumerate(contents):
            connector = '└── ' if i == len(contents) - 1 else '├── '
            print(f'{prefix}{connector}{item.name}')
            
            if item.is_dir():
                extension = '    ' if i == len(contents) - 1 else '│   '
                self.print_tree(path=item, prefix=prefix+extension)


import os
from os.path import join as _j
import sqlite3
from typing import List
import functools
import pandas as pd
from datetime import datetime
import logging
from pathlib import Path

from bellastore.filesystem.fs import Fs
from bellastore.utils.scan import Scan

# DATABSES
def sqlite_connection(func):
    ''' 
    Decorator allowing for comprehensive sqlite connection
    '''
    
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        with sqlite3.connect(self.sqlite_path) as conn:
            cursor = conn.cursor()
            try:
                result = func(self, cursor, *args, **kwargs)
                conn.commit()
                return result
            except Exception as e:
                conn.rollback()
                raise e
    return wrapper

class Db(Fs):
    ''' 
    A class representing an sqlite databse abstracting the underlying file system

    Attributes
    ----------
    filename: str
        The name of the database, classically `scans.sqlite`
    sqlite_path:
        The path to the database.
        The databse is placed always on top level within the storage directory.
    
    Methods
    -------
    insert_from_ingress:
        The method for inserting several scans to the storage
    '''

    def __init__(self, root_dir, ingress_dir, filename):
        super().__init__(root_dir, ingress_dir)
        self.filename = filename
        self.sqlite_path = os.path.join(self.storage_dir, self.filename)
        self._initialize_db()

    @sqlite_connection
    def _table_exists(self, cursor, table_name: str):
        cursor.execute(
            "SELECT name FROM sqlite_master WHERE name=?", (table_name, ))
        result = cursor.fetchall()
        if result:
            return True
        else:
            False

    @sqlite_connection
    def _initialize_db(self, cursor):
        if not self._table_exists('ingress'):
            cursor.execute('''
            CREATE TABLE ingress (
                hash TEXT,
                filepath TEXT,
                filename TEXT,
                UNIQUE(hash, filepath, filename)
            )
            ''')
        if not self._table_exists('storage'):
            cursor.execute('''
            CREATE TABLE storage (
                hash TEXT NOT NULL PRIMARY KEY,
                filepath TEXT,
                filename TEXT,
                scanname TEXT,
                FOREIGN KEY(hash) REFERENCES ingress(hash)
            )
            ''')

    @sqlite_connection  
    def add_scan_to_ingress_db(self, cursor, scan: Scan, rec = False):
        # rec tells us if the scan is already recorded in storage, so we do not
        # have to physically move the file around
        if not rec:
            self._add_scan_to_ingress(scan)
        print(f"Recording scan in ingress")
        cursor.execute(
            f"INSERT INTO ingress (hash, filepath, filename) VALUES (?, ?, ?)",
            (scan.hash, scan.path, scan.filename)
        )
    def add_scans_to_ingress_db(self, scans: List[Scan]):
        for scan in scans:
            self.add_scan_to_ingress_db(scan)

    @sqlite_connection  
    def add_scan_to_storage_db(self, cursor, scan: List[Scan]):
        self.add_scan_to_ingress_db(scan)
        # This is super important
        self.add_scan_to_storage(scan)
        print(f"Recording scan in storage")
        cursor.execute(f"""
            INSERT INTO storage (hash, filepath, filename, scanname) 
            VALUES (?, ?, ?, ?)
            """, (scan.hash, scan.path, scan.filename, scan.scanname))

    def add_scans_to_storage_db(self, scans: List[Scan]):
        for scan in scans:
            self.add_scan_to_storage_db(scan)

    def insert(self, scan: Scan):
        ''' 
        Inserts a single scan into the storage database.

        Inserting into the storage follows multiple steps:
        - check wether scan is already recorded in ingress
        - record scan in storage db (if not existent)
        - move scan file to storage (if not existent)
        '''
        scan.hash_scan()
        print(f"\nStarting insert pipeline for {scan.hash}:")
        existing_ingress_entries = self.get_entries_from_ingress_db()
        if (scan.hash, scan.path, scan.scanname) in existing_ingress_entries:
            print(f'Scan already recorded in the ingress table and thus scan will be deleted.')
            os.remove(scan.path)
            scan.path = None
            return
        existing_storage_entries = self.get_entries_from_storage_db()
        existing_storage_hashes = [entry[0] for entry in existing_storage_entries]
        if scan.hash in existing_storage_hashes:
            print(f'Scan already recorded in the storage table, so scan will be only recorded in ingress table and then deleted.')
            self.add_scan_to_ingress_db(scan, rec = True)
            print(f'Deleting scan')
            os.remove(scan.path)
            scan.path = None
            return
        # In this case the scan is not recorded, thus also not in storage so we run the whole pipeline
        # This will also automatically care about the storage backend recursively
        # TODO: the scan is already hashed at this place, so we do not need to hash it again
        self.add_scan_to_storage_db(scan)
    
    def insert_many(self, scans: List[Scan]):
        for scan in scans:
            self.insert(scan)
    
    def insert_from_ingress(self):
        ''' This is the main insert function

        This function retrieves valid scans from the ingress, inserts
        those into the storage and removes the resulting empty folders from the ingress.

        Return
        ------
            Valid scans (no matter if already in storage or not)
        '''
        scans = self.get_valid_scans_from_ingress()
        self.insert_many(scans)
        self.remove_empty_folders()
        return scans



    @sqlite_connection 
    def _read_all(self, cursor, table_name: str):
        cursor.execute(f"SELECT * FROM {table_name}")
        data = cursor.fetchall()
        return data
      
    def _read_all_pd(self, table_name: str):
        conn = sqlite3.connect(self.sqlite_path)
        df = pd.read_sql_query(f"SELECT * FROM {table_name}", conn)
        conn.close()
        return df


    def get_entries_from_ingress_db(self):
        return self._read_all('ingress')
    
    def get_entries_from_storage_db(self):
        return self._read_all('storage')
    
    def setup_logging(self):
        """Configure logging to both file and console"""
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(_j(self.backup_dir, 'database_backup.log')),
                logging.StreamHandler()
            ]
        )
    

    def create_backup(self, max_backups=10):
        """
        Create a backup of the SQLite database
        
        Args:
            max_backups (int): Maximum number of backup files to keep
        """
        try:
            
            # Generate backup filename with timestamp
            timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
            db_name = Path(self.sqlite_path).stem
            backup_path = Path(self.backup_dir) / f"{db_name}_backup_{timestamp}.sqlite"
            
            # Ensure source database exists
            if not Path(self.sqlite_path).exists():
                raise FileNotFoundError(f"Database file not found: {self.sqlite_path}")
            
            # Create backup using SQLite's backup API
            with sqlite3.connect(self.sqlite_path) as source:
                with sqlite3.connect(str(backup_path)) as target:
                    source.backup(target)
            
            logging.info(f"Backup created successfully: {backup_path}")
            
            # Clean up old backups if exceeding max_backups
            self._cleanup_old_backups(max_backups)
            
            return True
            
        except Exception as e:
            logging.error(f"Backup failed: {str(e)}")
            return False

    def _cleanup_old_backups(self, max_backups):
        """Remove oldest backups if exceeding max_backups limit"""
        backups = sorted(
            Path(self.backup_dir).glob('*_backup_*.sqlite'),
            key=os.path.getctime
        )
        
        while len(backups) > max_backups:
            oldest_backup = backups.pop(0)
            oldest_backup.unlink()
            logging.info(f"Removed old backup: {oldest_backup}")    

    
    def __str__(self):
        # TODO: print_tree method should return a string to be passed to the return
        print("\n")
        self.print_tree()
        ingress_df = self._read_all_pd('ingress')
        storage_df = self._read_all_pd('storage')
        return(f"Ingress DB:\n {ingress_df.to_string()}\n Storage DB:\n {storage_df.to_string()}\n")


from typing import List

# TODO: currently .mrxs is not supported, this needs some implementation in the Scans.move function

scan_extensions         : List[str]     = [".ndpi", ".svs",  ".tif", ".tiff"]        # Used in `scan.py` for simply checking via `endswith` if the file is valid
scan_extensions_glob    : List[str]     = ["*.ndpi", "*.svs", "*.tif", "*.tiff"]    # Used in `scan_database.py` to find all scanner files using globs


import os
from os.path import join as _j
import sqlite3

from bellastore.database.db import Db


def table_exists(sqlite_path, table_name: str):
    with sqlite3.connect(sqlite_path) as conn:
        cursor = conn.cursor()
        cursor.execute(
            "SELECT name FROM sqlite_master WHERE name=?", (table_name, ))
        result = cursor.fetchall()
    if result:
        return True
    else:
        False

def check_tables_exists(sqlite_path):
    for table_name in ['ingress', 'storage']:
        table_exists(sqlite_path, table_name)


def test_initialization(root_dir, ingress_dir):
    db = Db(root_dir, ingress_dir, 'scans.sqlite')
    check_tables_exists(db.sqlite_path)

    